#pragma once

#include "registers.h"
#include <stdint.h>

namespace drivers {
namespace uart_16550 {

enum class DataBits {
    BITS_5,
    BITS_6,
    BITS_7,
    BITS_8,
};

enum class StopBits {
    STOP_1,
    STOP_1_5,
    STOP_2,
};

enum class Parity {
    NONE,
    EVEN,
    ODD,
};

struct Uart16550 {
    Uart16550(void * const addr) :
        m_registers(*reinterpret_cast<Registers *>(addr)) {
    }

    void init(
            const int baud_rate,
            const DataBits data_bits,
            const Parity parity,
            const StopBits stop_bits) const volatile {
        switch(data_bits) {
            case DataBits::BITS_5:
                m_registers.lcr.word_length = 0;
                break;
            case DataBits::BITS_6:
                m_registers.lcr.word_length = 1;
                break;
            case DataBits::BITS_7:
                m_registers.lcr.word_length = 2;
                break;
            case DataBits::BITS_8:
                m_registers.lcr.word_length = 3;
                break;
        }

        switch(parity) {
            case Parity::NONE:
                m_registers.lcr.parity_enable = 0;
                break;
            case Parity::EVEN:
                m_registers.lcr.parity_enable = 1;
                m_registers.lcr.even_parity = 1;
                break;
            case Parity::ODD:
                m_registers.lcr.parity_enable = 1;
                m_registers.lcr.even_parity = 0;
                break;
        }

        switch(stop_bits) {
            case StopBits::STOP_1:
                m_registers.lcr.stop_bits = 0;
                break;
            case StopBits::STOP_1_5:
                // TODO: Throw error if data_bits isnt BITS_5
                m_registers.lcr.stop_bits = 0;
                break;
            case StopBits::STOP_2:
                // TODO: Throw error if data_bits is BITS_5
                m_registers.lcr.stop_bits = 1;
                break;
        }

        // Assert DLAB to allow setting the divisor
        // TODO: Assert CHCFG_AT_BUSY (UART_HALT[1])?
        m_registers.lcr.dlab = 1;

        // TODO: Get actual clock from TCC
        const uint16_t divisor = 24000000 / baud_rate / 16;
        m_registers.dch = divisor >> 8;
        m_registers.dcl = divisor & 0xFF;

        // TODO: Assert CHANGE_UPDATE (UART_HALT[2])?
        // TODO: Wait for CHANGE_UPDATE (UART_HALT[2]) to be deasserted?

        // Deassert DLAB to return to normal operation
        m_registers.lcr.dlab = 0;

        // Disable interrupts
        // TODO: Move to its own function
        // TODO: Cheese?
        reinterpret_cast<volatile uint8_t &>(m_registers.ier) = 0;

        // Enable FIFO
        // TODO: Move to its own function
        m_registers.fcr.fifo_enable = 1;
    }

    void write(const char c) {
        while(m_registers.lsr.thr_empty == 0);
        m_registers.thr = c;
    }

  private:
    volatile Registers &m_registers;
};

}  // End namespace uart_16550
}  // End namespace drivers
